/**
 * Munich University of Applied Sciences, FB07 Computer Science
 * Entscheidungstheorie - Prof.Dr.Peters - Master of Science
 * 
 * System: Windows XP, Linux 2.6, Sun Java JDK 6.0
 * Semestergroup: IG2
 */

/**
 * USABILITY:
 *  Copyright (c) 2006
 *  Thomas Woellert (Dipl.-Inf. (FH)) (thomas@woellert.net)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

package net.woellert.viergewinnt;

import java.io.File;
import java.io.FileInputStream;
import java.util.Properties;

import net.woellert.viergewinnt.constants.ApplicationConstants;
import net.woellert.viergewinnt.objects.Game;
import net.woellert.viergewinnt.objects.Playfield;
import net.woellert.viergewinnt.objects.PlayfieldFrame;
import net.woellert.viergewinnt.player.AbstractPlayer;
import net.woellert.viergewinnt.player.CPUPlayer;
import net.woellert.viergewinnt.player.ExternalPlayer;
import net.woellert.viergewinnt.player.HumanPlayer;
import net.woellert.viergewinnt.util.logging.LoggingManager;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;

/**
 * Class Description:
 * 		Main startup class
 * 
 * @author Thomas Woellert (thomas@woellert.net)
 */
public class VierGewinntStartup {

	/* Our local logger object: Null at the beginning, because we need to initialize the LoggingManager first */
	protected static Logger localLogger = null;

	// The main frame
	protected static PlayfieldFrame mainFrame = null;
	
	// The current Player
	public static int CURRENT_PLAYER = 1;
	
	// The players
	public static AbstractPlayer PLAYER_1 = null;
	public static AbstractPlayer PLAYER_2 = null;
	
	/**
	 * Constructor
	 * @param player1 the first player (starts first)
	 * @param player2 the second player
	 */
	public VierGewinntStartup(AbstractPlayer player1, AbstractPlayer player2) {
		
		// Remember the players
		VierGewinntStartup.PLAYER_1 = player1;
		VierGewinntStartup.PLAYER_2 = player2;
		
		// Do check
		if(player1 == null) {
			if (localLogger.isEnabledFor(Level.ERROR)) {
				localLogger.log(Level.ERROR, "Ooops, something went wrong during creation of player 1. Check your input arguments !");
			}
			System.out.println(ApplicationConstants.getHelpText());
			System.exit(0);
			
		} else if(player2 == null) {
			if (localLogger.isEnabledFor(Level.ERROR)) {
				localLogger.log(Level.ERROR, "Ooops, something went wrong during creation of player 2. Check your input arguments !");
			}
			System.out.println(ApplicationConstants.getHelpText());
			System.exit(0);
		}

		// Create the playfield
		Playfield field = new Playfield(ApplicationConstants.PLAYFIELD_SIZE_ROWS, ApplicationConstants.PLAYFIELD_SIZE_COLS);

		// Display the playfield graphically in a panel
		mainFrame = field.getFrame();
		mainFrame.setVisible(true);

		// Run the game
		Game game = new Game(player1, player2, field);
		game.run();
	}

	/**
	 * Start method
	 * @param args the input arguments
	 */
	protected static void start(String[] args) {

		// Check if the min number of input arguments have been given
		if (args.length < 5) {
			System.out.println(ApplicationConstants.getHelpText());
			System.exit(0);
		}
		
		// Grab the arguments #1 #2 and #3 defining the playfield and winning combination
		int cols = -1;
                int rows = -1;
                int wingroup = -1;
                
                try {
	                cols = Integer.parseInt(args[0]);
	                rows = Integer.parseInt(args[1]);
	                wingroup = Integer.parseInt(args[2]);
	                
                } catch (NumberFormatException e) {}
		
		if(cols >= 1) {
			ApplicationConstants.PLAYFIELD_SIZE_COLS = cols;
		} else {
			if (localLogger.isEnabledFor(Level.ERROR)) {
				localLogger.log(Level.ERROR, "No valid number of columns given: " + args[0]);
			}
			System.out.println(ApplicationConstants.getHelpText());
			System.exit(0);
		}
		
		if(rows >= 1) {
			ApplicationConstants.PLAYFIELD_SIZE_ROWS = rows;
		} else {
			if (localLogger.isEnabledFor(Level.ERROR)) {
				localLogger.log(Level.ERROR, "No valid number of rows given: " + args[1]);
			}
			System.out.println(ApplicationConstants.getHelpText());
			System.exit(0);
		}
		
		if(wingroup >= 1) {
			ApplicationConstants.PLAYFIELD_WIN_SIZE = wingroup;
		} else {
			if (localLogger.isEnabledFor(Level.ERROR)) {
				localLogger.log(Level.ERROR, "No valid win size given: " + args[2]);
			}
			System.out.println(ApplicationConstants.getHelpText());
			System.exit(0);
		}
		
		// Info
		if (localLogger.isEnabledFor(Level.INFO)) {
			localLogger.log(Level.INFO, "Playfield Properties: ");
			localLogger.log(Level.INFO, "	" + ApplicationConstants.PLAYFIELD_SIZE_ROWS + " rows");
			localLogger.log(Level.INFO, "	" + ApplicationConstants.PLAYFIELD_SIZE_COLS + " cols");
			localLogger.log(Level.INFO, "	" + ApplicationConstants.PLAYFIELD_WIN_SIZE + " in a row needed");
		}

		// Grab arguments #4 and #5 defining which players to use
		int inputArgumentIndex = 5;
		AbstractPlayer player1 = null;
		AbstractPlayer player2 = null;

		// Create Player 1
		if (args[3].charAt(0) == 'c') {
			
			if (localLogger.isEnabledFor(Level.INFO)) {
				localLogger.log(Level.INFO, "Player 1: CPU");
			}
			
			// CPU Player, with additional information concerning the search tree depth to use
			if(inputArgumentIndex >= args.length) {
				if (localLogger.isEnabledFor(Level.ERROR)) {
					localLogger.log(Level.ERROR, "Not enough input arguments given !");
				}
				System.out.println(ApplicationConstants.getHelpText());
				System.exit(1);
			}
			
			int maxTreeDepth = 2;
			
			try {
	                        maxTreeDepth = Integer.parseInt(args[inputArgumentIndex]);
	                        inputArgumentIndex++;
                        } catch (NumberFormatException e) {
                        	if (localLogger.isEnabledFor(Level.ERROR)) {
					localLogger.log(Level.ERROR, "Wrong input argument given (no number) !");
				}
                        	System.out.println(ApplicationConstants.getHelpText());
                        	System.exit(0);
                        }
			
			if(maxTreeDepth < 1) {
				if (localLogger.isEnabledFor(Level.ERROR)) {
					localLogger.log(Level.ERROR, "Search Tree Depth of CPU Player MUST be greater than or equal to 1 !");
				}
				System.exit(0);
			}
			
			player1 = new CPUPlayer(maxTreeDepth);
			
			if (localLogger.isEnabledFor(Level.INFO)) {
				localLogger.log(Level.INFO, "Search Tree Depth: " + maxTreeDepth);
			}
			
		} else if (args[3].charAt(0) == 'h') {
			
			if (localLogger.isEnabledFor(Level.INFO)) {
				localLogger.log(Level.INFO, "Player 1: Human");
			}
			
			// Human Player
			player1 = new HumanPlayer();
			
		} else if (args[3].charAt(0) == 'e') {
			
			// Check if dat file argument is present
			if(args.length <= inputArgumentIndex) {
				if (localLogger.isEnabledFor(Level.ERROR)) {
					localLogger.log(Level.ERROR, "You need to provide a DAT file to use for the 1st Player !");
				}
				System.out.println(ApplicationConstants.getHelpText());
				System.exit(0);
			}
			
			// External Player
			player1 = new ExternalPlayer(new File(args[inputArgumentIndex]));
			inputArgumentIndex++;
		}

		// Create Player 2
		if (args[4].charAt(0) == 'c') {
			
			if (localLogger.isEnabledFor(Level.INFO)) {
				localLogger.log(Level.INFO, "Player 2: CPU");
			}
			
			// CPU Player, with additional information concerning the search tree depth to use
			if(inputArgumentIndex >= args.length) {
				if (localLogger.isEnabledFor(Level.ERROR)) {
					localLogger.log(Level.ERROR, "Not enough input arguments given !");
				}
				System.out.println(ApplicationConstants.getHelpText());
				System.exit(0);
			}
			
			int maxTreeDepth = 2;
			
			try {
	                        maxTreeDepth = Integer.parseInt(args[inputArgumentIndex]);
	                        inputArgumentIndex++;
                        } catch (NumberFormatException e) {
                        	if (localLogger.isEnabledFor(Level.ERROR)) {
					localLogger.log(Level.ERROR, "Wrong input argument given (no number) !");
				}
                        	System.out.println(ApplicationConstants.getHelpText());
                        	System.exit(0);
                        }
			
			if(maxTreeDepth < 1) {
				if (localLogger.isEnabledFor(Level.ERROR)) {
					localLogger.log(Level.ERROR, "Search Tree Depth of CPU Player MUST be greater than or equal to 1 !");
				}
				System.exit(1);
			}
			
			player2 = new CPUPlayer(maxTreeDepth);
			
			if (localLogger.isEnabledFor(Level.INFO)) {
				localLogger.log(Level.INFO, "Search Tree Depth: " + maxTreeDepth);
			}
			
		} else if (args[4].charAt(0) == 'h') {
			
			if (localLogger.isEnabledFor(Level.INFO)) {
				localLogger.log(Level.INFO, "Player 2: Human");
			}
			
			// Human Player
			player2 = new HumanPlayer();
			
		} else if (args[4].charAt(0) == 'e') {
			
			// Check if dat file argument is present
			if(args.length <= inputArgumentIndex) {
				if (localLogger.isEnabledFor(Level.ERROR)) {
					localLogger.log(Level.ERROR, "You need to provide a DAT file to use for the 2nd Player !");
				}
				System.out.println(ApplicationConstants.getHelpText());
				System.exit(0);
			}
			
			// External Player
			player2 = new ExternalPlayer(new File(args[inputArgumentIndex]));
			inputArgumentIndex++;
		}

		// Start the game
		new VierGewinntStartup(player1, player2);
	}

	/**
	 * Call this method to start the program
	 * @param args, command line arguments
	 */
	public static void main(String[] args) {

		try {
			// Load the main application properties
			Properties propsIn = new Properties();

			try {
				FileInputStream inputStream = new FileInputStream(ApplicationConstants.GLOBAL_PROPS);
				propsIn.load(inputStream);
			} catch (Exception e) {
				System.err.println("Failed to load properties file: " + ApplicationConstants.GLOBAL_PROPS);
			}

			// Initialize the logging manager
			LoggingManager.setProperties(propsIn);

			// Now we can initialize our local logger
			localLogger = LoggingManager.getLogger(VierGewinntStartup.class.getName());

			// Check if we properly received a logger
			if (localLogger != null) {

				if (localLogger.isEnabledFor(Level.INFO)) {
					// Print startup text
					System.out.println(ApplicationConstants.getStartupText());
					localLogger.log(Level.INFO, "Starting VierGewinnt ...");
				}

				// Separate the command-line arguments and start the application
				VierGewinntStartup.start(args);

			} else {
				System.err.println("[FATAL] - LoggingManager and/or Logger initialization failed - Aborting !");
			}

		} catch (Exception e) {

			if (localLogger != null) {
				if (localLogger.isEnabledFor(Level.ERROR)) {
					localLogger.log(Level.ERROR, "Running Startup failed!", e);
				}
			} else {
				System.err.println("[FATAL] - Running Startup failed !");
				e.printStackTrace();
			}
		}
	}
	
	/**
	 * Should be called if the game is over
	 * @param winner the winner (1 = Player 1, 2 = Player 2)
	 */
	public static void gameIsOver(int winner) {
		
		// Check for even
		if(winner == 0) {
			mainFrame.setTitle(ApplicationConstants.FRAME_TITLE + " - No WINNER");
		} else {
			mainFrame.setTitle(ApplicationConstants.FRAME_TITLE + " - WINNER - Player " + winner);
		}
		
		CURRENT_PLAYER = 0;
	}
	
	/**
	 * Should be called to announce the current player in the title bar
	 * @param player the player (1 = Player 1, 2 = Player 2)
	 */
	public static void currentPlayer(int player) {
		CURRENT_PLAYER = player;
		mainFrame.setTitle(ApplicationConstants.FRAME_TITLE + " - Player " + player + " your turn ...");
		
		// Also repaint the playfield
		mainFrame.repaint();
	}
	

	/**
	 * Retrieve a player
	 * @return the player
	 */
	public static AbstractPlayer getCurrentPlayer() {
        	if(CURRENT_PLAYER == 1) {
        		return PLAYER_1;
        	} else if(CURRENT_PLAYER == 2) {
        		return PLAYER_2;
        	} else {
        		return null;
        	}
        }	
}
