/**
 * Munich University of Applied Sciences, FB07 Computer Science
 * Entscheidungstheorie - Prof.Dr.Peters - Master of Science
 * 
 * System: Windows XP, Linux 2.6, Sun Java JDK 6.0
 * Semestergroup: IG2
 */

/**
 * USABILITY:
 *  Copyright (c) 2006
 *  Thomas Woellert (Dipl.-Inf. (FH)) (thomas@woellert.net)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

package net.woellert.viergewinnt.objects;

import java.util.ArrayList;

import net.woellert.viergewinnt.constants.ApplicationConstants;

/**
 * Class Description:
 * 		The playfield
 * 
 * @author Thomas Woellert (thomas@woellert.net)
 */
public class Playfield {

	// The number of rows and columns
	protected int rows = -1;
	protected int cols = -1;

	// The playfield
	protected SlotState[][] field = null;

	// The change listener
	protected PlayfieldChangeListener listener = null;

	// The frame displaying the playfield
	protected PlayfieldFrame frame = null;

	// The last turn
	protected int lastTurnCol = -1;
	protected int lastTurnRow = -1;

	/**
	 * SlotStates
	 */
	public enum SlotState {
		EMPTY, PLAYER_1, PLAYER_2,
	}

	/**
	 * Constructor
	 * @param rows the number of rows
	 * @param cols the number of columns
	 */
	public Playfield(int rows, int cols) {
		this.rows = rows;
		this.cols = cols;

		// Create the field
		this.field = new SlotState[rows][cols];

		// Clear the field
		for (int i = 0; i < rows; i++) {
			for (int j = 0; j < cols; j++) {
				this.field[i][j] = SlotState.EMPTY;
			}
		}
	}

	/**
	 * Copy Constructor
	 * @param f the playfield to copy
	 */
	public Playfield(Playfield f) {
		this.rows = f.getRows();
		this.cols = f.getCols();

		// Create the empty field
		this.field = new SlotState[rows][cols];

		// Copy the field
		for (int i = 0; i < rows; i++) {
			for (int j = 0; j < cols; j++) {
				this.field[i][j] = f.getState(i, j);
			}
		}
	}

	/**
	 * Copy a new playfield into the existing one
	 * @param field
	 */
	public void newPlayField(Playfield field) {

		for (int i = 0; i < field.getRows(); i++) {
			for (int j = 0; j < field.getCols(); j++) {
				this.setState(i, j, field.getState(i, j));
			}
		}
	}

	/**
	 * Retrieve the frame displaying the playfield
	 * @return the frame
	 */
	public PlayfieldFrame getFrame() {
		if (frame == null) {
			frame = new PlayfieldFrame(this);
		}

		return frame;
	}

	/**
	 * Get the most bottom empty row for this column
	 * @param col the column 
	 * @return the most bottom empty row
	 */
	public int getMostBottomRow(int col) {

		int mostBottomRow = -1;

		for (int i = (this.getRows() - 1); i >= 0; i--) {

			if (this.getState(i, col) == SlotState.EMPTY) {
				mostBottomRow = i;
			}
		}

		return mostBottomRow;
	}

	/**
	 * Retrieve all currently legal turns
	 * @return the array with all currently legal columns
	 */
	public int[] getAllLegalTurns() {

		ArrayList<Integer> legalCols = new ArrayList<Integer>();

		for (int i = 0; i < cols; i++) {
			int mostBottomRow = getMostBottomRow(i);

			if (mostBottomRow != -1) {
				legalCols.add(i);
			}
		}

		int[] allLegalTurns = new int[legalCols.size()];

		for (int i = 0; i < legalCols.size(); i++) {
			allLegalTurns[i] = legalCols.get(i);
		}

		return allLegalTurns;
	}

	/**
	 * Check if we have a winner
	 * @return 1 for player1, 2 for player2, -1 for even
	 */
	public int isGameOver() {

		int winner = 0;

		// Check for 4 in a row			
		for (int i = 0; i < rows; i++) {
			SlotState[] currentRow = new SlotState[ApplicationConstants.PLAYFIELD_SIZE_COLS];

			for (int j = 0; j < cols; j++) {
				currentRow[j] = this.field[i][j];
			}

			winner = checkArrayForWinner(currentRow);

			if (winner != 0) {
				return winner;
			}
		}

		// Check for 4 in a column
		for (int j = 0; j < cols; j++) {
			SlotState[] currentCol = new SlotState[ApplicationConstants.PLAYFIELD_SIZE_ROWS];

			for (int i = 0; i < rows; i++) {
				currentCol[i] = this.field[i][j];
			}

			winner = checkArrayForWinner(currentCol);

			if (winner != 0) {
				return winner;
			}
		}

		// Check for 4 in a diagonal		
		for (int startRow = 0; startRow < ApplicationConstants.PLAYFIELD_SIZE_ROWS; startRow++) {
			SlotState[] diagRaising = getDiagonalRaising(startRow, 0);
			winner = checkArrayForWinner(diagRaising);

			if (winner != 0) {
				return winner;
			}

			SlotState[] diagFalling = getDiagonalFalling(startRow, 0);
			winner = checkArrayForWinner(diagFalling);

			if (winner != 0) {
				return winner;
			}
		}

		for (int startCol = 0; startCol < ApplicationConstants.PLAYFIELD_SIZE_COLS; startCol++) {
			SlotState[] diagRaising = getDiagonalRaising(0, startCol);
			winner = checkArrayForWinner(diagRaising);

			if (winner != 0) {
				return winner;
			}

			SlotState[] diagFalling = getDiagonalFalling(ApplicationConstants.PLAYFIELD_SIZE_ROWS-1, startCol);
			winner = checkArrayForWinner(diagFalling);

			if (winner != 0) {
				return winner;
			}
		}

		// Check if the game is over because no free slots are left
		int freeSlotsCounter = 0;

		for (int i = 0; i < rows; i++) {
			for (int j = 0; j < cols; j++) {
				if (field[i][j] == SlotState.EMPTY) {
					freeSlotsCounter++;
				}
			}
		}

		if (freeSlotsCounter == 0) {
			return -1;
		}

		return winner;
	}

	/**
	 * Checks the given array if we have a winner
	 * @param states
	 * @return 1 for player1, 2 for player2, -1 for even
	 */
	public int checkArrayForWinner(SlotState[] states) {

		SlotState prevState = SlotState.EMPTY;
		int counter = 0;

		for (int i = 0; i < states.length; i++) {

			SlotState state = states[i];

			if (prevState == SlotState.EMPTY) {

				if (state == SlotState.PLAYER_1 || state == SlotState.PLAYER_2) {
					prevState = state;
					counter = 1;

				} else if (state == SlotState.EMPTY) {
					counter = 0;
				}

			} else if (prevState == SlotState.PLAYER_1) {

				if (state == SlotState.PLAYER_1) {
					counter++;

					if (counter == ApplicationConstants.PLAYFIELD_WIN_SIZE) {
						// Player 1 won with enough items in a row
						return 1;
					}

				} else if (state == SlotState.PLAYER_2) {
					prevState = state;
					counter = 1;

				} else if (state == SlotState.EMPTY) {
					prevState = state;
					counter = 0;
				}

			} else if (prevState == SlotState.PLAYER_2) {

				if (state == SlotState.PLAYER_1) {
					prevState = state;
					counter = 1;

				} else if (state == SlotState.PLAYER_2) {
					counter++;

					if (counter == ApplicationConstants.PLAYFIELD_WIN_SIZE) {
						// Player 2 won with enough items in a row
						return 2;
					}

				} else if (state == SlotState.EMPTY) {
					prevState = state;
					counter = 0;
				}
			}
		}

		return 0;
	}

	/**
	 * Get the raising diagonal "/" as array
	 * @param startRow the start row
	 * @param startCol the start col
	 * @return the array
	 */
	public SlotState[] getDiagonalRaising(int startRow, int startCol) {

		ArrayList<SlotState> states = new ArrayList<SlotState>();

		while (startRow < ApplicationConstants.PLAYFIELD_SIZE_ROWS && startCol < ApplicationConstants.PLAYFIELD_SIZE_COLS) {

			states.add(field[startRow][startCol]);
			startRow++;
			startCol++;
		}

		SlotState[] statesArray = new SlotState[states.size()];

		for (int i = 0; i < states.size(); i++) {
			statesArray[i] = states.get(i);
		}

		return statesArray;
	}

	/**
	 * Get the falling diagonal "\" as array
	 * @param startRow the start row
	 * @param startCol the start col
	 * @return the array
	 */
	public SlotState[] getDiagonalFalling(int startRow, int startCol) {

		ArrayList<SlotState> states = new ArrayList<SlotState>();

		while (startRow < ApplicationConstants.PLAYFIELD_SIZE_ROWS && startCol < ApplicationConstants.PLAYFIELD_SIZE_COLS && startRow >= 0) {
			states.add(field[startRow][startCol]);
			startRow--;
			startCol++;
		}

		SlotState[] statesArray = new SlotState[states.size()];

		for (int i = 0; i < states.size(); i++) {
			statesArray[i] = states.get(i);
		}

		return statesArray;
	}

	/**
	 * Do a turn (throw the coin into the defined column)
	 * @param col the column
	 * @param state the player coin to use
	 */
	public void doTurn(int col, SlotState state) {

		for (int i = 0; i < rows; i++) {
			if (this.field[i][col] == SlotState.EMPTY) {
				lastTurnRow = i;
				lastTurnCol = col;
				setState(i, col, state);
				break;
			}
		}
	}

	/**
	 * Undo the last turn
	 */
	public void undoTurn() {

		if (lastTurnRow == -1 || lastTurnCol == -1) {
			return;
		}

		setState(lastTurnRow, lastTurnCol, SlotState.EMPTY);

		lastTurnRow = -1;
		lastTurnCol = -1;
	}

	/**
	 * Manually set a playfield items state
	 * @param row the row
	 * @param col the col
	 * @param state the state to set
	 */
	public void setState(int row, int col, SlotState state) {
	        this.field[row][col] = state;
               
		if (this.listener != null) {
			this.listener.playfieldChanged(this);
		}
	}

	/**
	 * Retrieve a playfield items state
	 * @param row the row
	 * @param col the col
	 * @return the state
	 */
	public SlotState getState(int row, int col) {
		return field[row][col];
	}

	/**
	 * Get a numerical representation of the items state
	 * @param row the row
	 * @param col the col
	 * @return 1 for player1, -1 for player2, 0 for empty
	 */
	public int getStateNum(int row, int col) {

		if (row >= rows || col >= cols || row < 0 || col < 0) {
			return Integer.MIN_VALUE;
		}

		if (field[row][col] == SlotState.PLAYER_1) {
			return 1;
		} else if (field[row][col] == SlotState.PLAYER_2) {
			return -1;
		} else {
			return 0;
		}
	}

	/**
	 * Get the number of columns
	 * @return
	 */
	public int getCols() {
		return cols;
	}

	/**
	 * Get the number of rows
	 * @return
	 */
	public int getRows() {
		return rows;
	}

	/**
	 * Add a change listener
	 * @param listener
	 */
	public void addChangeListener(PlayfieldChangeListener listener) {
		this.listener = listener;
	}

	/**
	 * String representation of the playfield
	 * @return
	 */
	public String toString() {

		String ret = "[Playfield]\n";
		ret += "[rows = " + this.rows + " / cols = " + this.cols + "] \n";

		for (int i = (rows - 1); i >= 0; i--) {
			for (int j = 0; j < cols; j++) {
				SlotState state = this.field[i][j];

				switch (state) {
				case EMPTY:
					ret += "- \t";
					break;

				case PLAYER_1:
					ret += "1 \t";
					break;

				case PLAYER_2:
					ret += "2 \t";
					break;

				default:
					ret += "NA";
				}
			}

			ret += "\n";
		}

		ret += "[End of Playfield]";
		return ret;
	}
}
